import re
from datetime import datetime, timedelta
from functools import total_ordering

import numpy as np
import pandas as pd
import tzlocal
from icalendar import Calendar, Event
from pytz import timezone


@total_ordering
class OfficeHour:
    def __init__(self, s):
        self.s_orig = s
        self.day_idx, s = parse_day(s)

        if s.count('-') != 1:
            raise ValueError(f'doesnt contain unique `-`: {s}')
        time_start, time_end = s.split('-')
        self.time_start, name0 = parse_time(time_start)
        self.time_end, name1 = parse_time(time_end)

        # build name (remove name0 or name1 if empty string)
        name_tup = [name0, name1]
        name_tup = [s for s in name_tup if s]
        self.name = ' '.join(name_tup)

    def to_tuple(self):
        return self.day_idx, self.time_start, self.time_end, self.name

    def __hash__(self):
        return hash(self.to_tuple())

    def __str__(self):
        return f'OfficeHour({self.s_orig})'

    def __lt__(self, other):
        return self.to_tuple() < other.to_tuple()

    def __eq__(self, other):
        return self.to_tuple() == other.to_tuple()

    def intersects(self, other):
        if self.day_idx != other.day_idx:
            # different days, can't intersect
            return False

        if self < other:
            # self begins first (or at same time)
            return self.time_end > other.time_start
        else:
            # other begins first (or at same time)
            return other.time_end > self.time_start

    def get_event_kwargs(self, date_start, date_end, tz=None, **kwargs):
        """ gets weekly recurring event arguments, to be passed to Event

        Args:
            date_start (str): start date
            date_end (str): end date
            tz (str or timezone, optional): timezone. If not given, the local
                timezone is used.
            **kwargs: Additional parameters to be included in the event

        Returns:
            kwargs: dictionary to be unapcked into Event object

        Raises:
            AttributeError: event exceeds the maximum weekly repeats (53 weeks)
        """
        # Convert date_start, date_end to date objects
        date_start = pd.to_datetime(date_start).date()
        date_end = pd.to_datetime(date_end).date()

        # Move start date up to the correct weekday
        while date_start.weekday() != self.day_idx:
            date_start += timedelta(days=1)

        # Handle timezone (default to local time if not provided)
        if tz is None:
            tz = tzlocal.get_localzone()
        tz = timezone(str(tz))

        # build output starts / ends
        dtstart = datetime.combine(date_start, self.time_start)
        kwargs['dtstart'] = tz.localize(dtstart)
        dtend = datetime.combine(date_start, self.time_end)
        kwargs['dtend'] = tz.localize(dtend)

        # Compute the number of weekly repeats before the end date
        date = date_start
        for repeats in range(52):
            if date > date_end:
                break
            date = date + timedelta(weeks=1)
        else:
            raise AttributeError(f'> 1 yr event: {date_start} to {date_end}')

        kwargs['rrule'] = {'freq': 'weekly', 'count': repeats}

        return kwargs


def get_intersection_dict(oh_list):
    # order office hours (from earliest to latest in the day)
    _oh_list = [OfficeHour(oh) for oh in oh_list]
    idx_map = np.argsort(_oh_list)
    _oh_list.sort()

    # find intersections (initialize with reflexivity)
    oh_int_dict = {idx: [idx] for idx in range(len(oh_list))}
    for idx0, oh0 in enumerate(_oh_list):
        for _idx1, oh1 in enumerate(_oh_list[idx0 + 1:]):
            # idx1 is consistent with ordering of _oh_list
            idx1 = _idx1 + idx0 + 1
            if oh0.intersects(oh1):
                oh_int_dict[idx0].append(idx1)
                oh_int_dict[idx1].append(idx0)
            else:
                # oh0 is before oh1, if oh0 doesn't intersect oh1 it can't
                # intersect any which come after it in _oh_list, its sorted
                break

    # map from indexing of _oh_list back to given oh_list indexing
    oh_int_dict = {idx_map[k]: [idx_map[_v] for _v in v]
                   for k, v in oh_int_dict.items()}

    return oh_int_dict


def parse_day(day_str):
    """ extracts a day of week, as index, from a string

    Args:
        day_str (str): string containing some day of the week

    Returns:
        day_idx (int): 0 for Monday, 1 for Tuesday, ...
        day_str_clean (str): String with the day and nearby time removed
    """
    day_regexes = [
        r'\bmon(day)?s?\b',
        r'\btue(s(day)?)?s?\b',
        r'\bwed(nesday)?s?\b',
        r'\bthu(r(sday)?)?s?\b',
        r'\bfri(day)?s?\b',
        r'\bsat(urday)?s?\b',
        r'\bsun(day)?s?\b'
    ]

    found = [(i, re.search(pat, day_str, re.IGNORECASE))
             for i, pat in enumerate(day_regexes)]
    matches = [(i, m) for i, m in found if m]

    assert len(matches) == 1, \
        f'Expected one day of week in "{day_str}", found {len(matches)}'

    day_idx, match = matches[0]

    # Remove the matched day and any nearby time
    idx0, idx1 = match.span()
    day_str_clean = day_str[:idx0] + day_str[idx1:]
    day_str_clean = day_str_clean.strip(" ,:-")

    return day_idx, day_str_clean


def parse_time(s):
    """ converts to timedelta, from beginning of day to time_str

    Args:
        s (str): comes in one of two formats: "6:30 PM" or "4 AM"

    Returns:
        time (time): time of day
        s_clean (re.Match): input s, having time removed
    """
    # Match patterns for 12-hour and 24-hour unambiguous formats
    patterns = [
        ('%I:%M%p', re.compile(r'\d{1,2}:\d{2}\s*(?:AM|PM)', re.IGNORECASE)),
        ('%I%p', re.compile(r'\d{1,2}\s*(?:AM|PM)', re.IGNORECASE))
    ]

    for fmt, pattern in patterns:
        match_list = pattern.findall(s)
        match len(match_list):
            case 0:
                # no match found
                continue
            case 1:
                # unique match found
                match = pattern.search(s)
                s_match = match.group().replace(' ', '')
                time = datetime.strptime(s_match, fmt).time()

                # Remove the matched day and any nearby time
                idx0, idx1 = match.span()
                s_clean = s[:idx0] + s[idx1:]
                s_clean = s_clean.strip(" ,:-")

                return time, s_clean
            case _:
                raise ValueError(f'Multiple times found: {s}')

    raise ValueError(f'Ambiguous or invalid time string: "{s}"')


def build_calendar(oh_ta_dict, date_start, date_end, **kwargs):
    """  builds a calendar, a set of events, from oh_ta_dict

    Args:
        oh_ta_dict (dict): keys are OfficeHour, vals are lists of str (TA
            names)
        date_start (str): starting date for office hours for course
            (inclusive), see  get_event_kwargs()
        date_end (str): ending date for office hours for course (inclusive),
            see get_event_kwargs()=

    Returns:
        cal (Calendar): ready to be exported to ics format
    """
    cal = Calendar()
    for oh, ta_list in oh_ta_dict.items():
        if not ta_list:
            # skip oh slots without any TAs
            continue
        ta_list = [ta.capitalize() for ta in sorted(ta_list)]
        summary = oh.name + ': ' + ', '.join(sorted(ta_list))

        _kwargs = oh.get_event_kwargs(summary=summary,
                                      date_start=date_start,
                                      date_end=date_end,
                                      **kwargs)

        # build event with proper attributes of event,
        _kwargs = kwargs | _kwargs
        event = Event()
        for key, val in _kwargs.items():
            event.add(key, val)
        cal.add_component(event)

    return cal



import re
import warnings
from copy import copy
from itertools import chain

import numpy as np
from scipy.optimize import linear_sum_assignment

INVALID = -1

# scale of preferences to random noise added (to shuffle match order)
STD_SCALE_NOISE = .00001


def match(prefs, oh_per_ta, max_ta_per_oh=None, oh_int_dict=None, shuffle=True,
          ta_name_list=None, seed=0):
    """ matches TA to OH slot to maximize sum of prefs achieved

    Args:
        prefs (np.array): (num_ta, num_oh) preference scores for every
            combination of ta and oh.  nan where unavailable
        oh_per_ta (int): office hours assigned per ta
        max_ta_per_oh (int): maximum ta assigned to any particular oh
        oh_int_dict (dict): keys are oh index, values are a list of all oh
            indexes (including itself) which intersect (overlap).  See
            get_intersection_dict().  if None passed, each oh only overlaps
            itself
        shuffle (bool): toggles shuffling of tie breaking (seems to prefer
            earlier ta_idx)
        ta_name_list (list): str associated with each TA (e.g. email).  used for
            more informative error messages to end user (default is None, TA
            index used in error messages in this case)
        seed: given to shuffling

    Returns:
        oh_ta_match (list of lists): oh_ta_match[oh_idx] is a list of the
            index of all tas assigned particular oh_idx
    """
    def get_name(idx):
        if ta_name_list is None:
            return f'TA_{idx}'
        else:
            return ta_name_list[idx]

    # set invalid entries with low score
    prefs = copy(prefs)
    prefs[np.isnan(prefs)] = INVALID

    # init
    num_ta, num_oh = prefs.shape
    oh_ta_match = [list() for _ in range(num_oh)]

    if max_ta_per_oh is None:
        max_ta_per_oh = num_ta

    if oh_int_dict is None:
        # each office hour only intersects itself
        oh_int_dict = {idx: [idx] for idx in range(num_oh)}

    # init random number generator
    rng = np.random.default_rng(seed=seed)
    std_pref = np.nanstd(prefs.flatten())
    if std_pref == 0:
        std_pref = 1

    for match_idx in range(oh_per_ta):
        # build new _prefs and _oh_ta_match per availability remaining.
        # (i.e. if each oh spot has 2 open spots then each column repeated
        # twice and _oh_ta_match has two refs to each list in oh_ta_match)
        pref_list = list()
        _oh_ta_match = list()
        for oh_idx, ta_list in enumerate(oh_ta_match):
            num_ta_spots_left = max_ta_per_oh - len(ta_list)

            for _ in range(num_ta_spots_left):
                pref_list.append(prefs[:, oh_idx])
                _oh_ta_match.append(ta_list)

        if not pref_list:
            # no more TA spots left (will warn before returning)
            break
        _prefs = np.stack(pref_list, axis=1)

        if shuffle:
            # add some noise to shuffle assignment order (don't add noise to
            # invalid positions)
            bool_invalid = _prefs == INVALID
            c = STD_SCALE_NOISE / std_pref
            _prefs = _prefs + rng.standard_normal(_prefs.shape) * c
            _prefs[bool_invalid] = INVALID

        # match
        ta_idx, oh_idx = linear_sum_assignment(cost_matrix=_prefs,
                                               maximize=True)

        # record & validate matches
        for _ta_idx, _oh_idx in zip(ta_idx, oh_idx):
            if _prefs[_ta_idx, _oh_idx] == INVALID:
                raise RuntimeError(f'no remaining availability: {get_name(_ta_idx)}')
            _oh_ta_match[_oh_idx].append(_ta_idx)

            # get oh_idx (in original indexing, recall that _oh_ta_match may
            # have multiple references to the same office hours section)
            oh_idx = oh_ta_match.index(_oh_ta_match[_oh_idx])

            # mark this spot as invalid for this TA
            for idx in oh_int_dict[oh_idx]:
                prefs[_ta_idx, idx] = INVALID

    # count oh per ta
    _oh_per_ta, _ = np.histogram(list(chain.from_iterable(oh_ta_match)),
                                 bins=np.arange(-.5, num_ta + .5))
    for _ta_idx, n in enumerate(_oh_per_ta):
        if n != oh_per_ta:
            warnings.warn(f'only {n} OH slots assigned: {get_name(_ta_idx)}')

    return oh_ta_match


def get_scale(oh_list, scale_dict, verbose=True):
    """ associates each scaling factor to all matching office hours in oh_list

    Args:
        oh_list (list): list of strings, each is an office hours slot
        scale_dict (dict): keys are regex which match any relevant office
            hours, values are multiplicative factors to adjust preferences
            in these hours by
        verbose (bool): toggles command line output

    Returns:
        scale (np.array): scaling factor for every office hours slot
    """
    if verbose:
        print('\nScaling office hours preferences:')
    scale = np.ones(len(oh_list))
    for regex, mult in scale_dict.items():
        match_exists = False
        for oh_idx, oh in enumerate(oh_list):
            if re.search(regex, oh):
                # multiplier is applicable to this office hours slot
                scale[oh_idx] *= mult
                print(f'{oh} multiplied by {mult} (cumulative scale={scale[oh_idx]})')
                match_exists = True

        if not match_exists:
            warnings.warn(f'scale not applied, no office hours match:{regex}')
    return scale


def get_perc_max(oh_ta_match, prefs):
    """ computes percent of maximum score achieved per TA

    a perc_max value of .9 for a TA means that the matching achieved 90% of
    the maximum preferences score for a TA if we matched to optimize only
    this TAs preferences.

    Args:
        oh_ta_match (list of lists): oh_ta_match[oh_idx] is a list of the index
            of all tas assigned particular oh_idx
        prefs (np.array): (num_ta, num_oh) preference scores for every
            combination of ta and oh.  nan where unavailable

    Returns:
        perc_max (np.array): (num_ta) perc_max score for each TA
    """
    num_ta, num_oh = prefs.shape

    # count oh per ta in dict
    num_oh = np.zeros(num_ta, dtype=int)
    for ta_list in oh_ta_match:
        for ta in ta_list:
            num_oh[ta] += 1

    # compute max score possible (giving this TA their num_oh slots which
    # maximize preference)
    ta_max = np.empty(num_ta)
    for ta_idx, _pref in enumerate(prefs):
        _pref = _pref[~np.isnan(_pref)]
        _pref.sort()
        ta_max[ta_idx] = sum(_pref[-num_oh[ta_idx]:])

    # compute score achieved for each TA
    ta_achieve = np.zeros(num_ta)
    for oh, ta_list in enumerate(oh_ta_match):
        for ta in ta_list:
            ta_achieve[ta] += prefs[ta, oh]

    return ta_achieve / ta_max



import pandas as pd


def extract_csv(f_csv):
    df = pd.read_csv(f_csv)

    # drop duplicates email (take latest)
    df['Timestamp'] = pd.to_datetime(df['Timestamp'])
    df.sort_values('Timestamp', inplace=True)
    df.drop_duplicates(subset=df.columns[1], keep='last', inplace=True)

    # extract necessary info
    email_list = df.iloc[:, 1].to_list()
    name_list = df.iloc[:, 2].to_list()
    oh_list = df.columns[3:].to_list()
    prefs = df.iloc[:, 3:].values.astype(float)

    return prefs, email_list, name_list, oh_list
